import os
from io import BytesIO, StringIO
from tempfile import mkstemp

from docx import Document
from reportlab.lib import colors
from reportlab.lib.pagesizes import landscape, letter
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.platypus import PageBreak, Paragraph, SimpleDocTemplate, Table, TableStyle
from reportlab.platypus.doctemplate import LayoutError
from rest_framework import status
from rest_framework.renderers import BaseRenderer, TemplateHTMLRenderer
from rest_framework_csv.renderers import CSVRenderer
from tablib import Dataset

RESPONSE_ERROR = (
    "Response data is a %s, not a Dataset! "
    "Did you extend ExportMixin?"
)
PDF_COLUMNS_PER_PAGE = 9


class ExportBaseRenderer(BaseRenderer):
    """Renders Datasets using their built in export implementation.
    Only works with serializers that return Datasets as their data object.
    Uses a StringIO to capture the output of dataset.export('[format]')
    """
    def render(self, data, accepted_media_type=None, renderer_context=None):
        if 'response' in renderer_context:
            status_code = renderer_context['response'].status_code
            if not status.is_success(status_code):
                return "Error: %s" % data.get('detail', status_code)

        if not isinstance(data, Dataset):
            raise Exception(
                RESPONSE_ERROR % type(data).__name__
            )

        self.init_output()
        args = self.get_export_args(data)
        kwargs = self.get_export_kwargs(data, renderer_context)
        self.render_dataset(data, *args, **kwargs)
        return self.get_output()

    def render_dataset(self, data, *args, **kwargs):
        self.output.write(data.export(self.format))

    def init_output(self):
        self.output = StringIO()

    def get_output(self):
        return self.output.getvalue()

    def get_export_args(self, data):
        return [self.output]

    def get_export_kwargs(self, data, renderer_context):
        return {}


class ExportFileRenderer(ExportBaseRenderer):
    """Renderer for output formats that absolutely must use a file
    (i.e. Excel)
    """
    def init_output(self):
        file, filename = mkstemp(suffix='.' + self.format)
        self.filename = filename
        os.close(file)

    def render_dataset(self, data, *args, **kwargs):
        with open(self.filename, "wb") as fp:
            fp.write(data.export(self.format))

    def get_export_args(self, data):
        return [self.filename]

    def get_output(self):
        file = open(self.filename, 'rb')
        result = file.read()
        file.close()
        os.unlink(self.filename)
        return result


class ExportHTMLRenderer(TemplateHTMLRenderer, ExportBaseRenderer):
    media_type = "text/html"
    format = "html"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        table = ExportBaseRenderer.render(
            self, data, accepted_media_type, renderer_context,
        )

        return TemplateHTMLRenderer.render(
            self, {'table': table}, accepted_media_type, renderer_context,
        )

    def get_template_context(self, data, renderer_context):
        view = renderer_context['view']
        request = renderer_context['request']

        data['name'] = view.get_view_name()
        data['description'] = view.get_view_description(html=True)
        data['url'] = request.path.replace('.html', '')
        full_path = request.get_full_path()
        if '?' in full_path:
            data['url_params'] = full_path[full_path.index('?'):]
        data['available_formats'] = [
            cls.format for cls in view.renderer_classes
            if cls.format != 'html'
        ]

        if hasattr(view, 'get_template_context'):
            data.update(view.get_template_context(data))

        return data

    def get_export_kwargs(self, data, renderer_context):
        return {
            'classes': 'ui-table table-stripe',
            'na_rep': '',
        }


class ExportCSVRenderer(ExportBaseRenderer):
    """
    Renders data frame as CSV
    """
    media_type = "text/csv"
    format = "csv"

    def get_export_kwargs(self, data, renderer_context):
        return {'encoding': self.charset}


class ExportJSONRenderer(ExportBaseRenderer):
    """Renders dataset as JSON"""
    media_type = "application/json"
    format = "json"

    date_format_choices = {'epoch', 'iso'}
    default_date_format = 'iso'

    def get_export_kwargs(self, data, renderer_context):
        request = renderer_context['request']

        date_format = request.GET.get('date_format', '')
        if date_format not in self.date_format_choices:
            date_format = self.default_date_format

        return {
            'date_format': date_format,
        }


class ExportOpenXMLRenderer(ExportFileRenderer):
    """Renders dataset as Excel (.xlsx)"""
    media_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"  # noqa
    format = "xlsx"


class ExportExcelRenderer(ExportFileRenderer):
    """Renders dataset as Excel (.xls)"""
    media_type = "application/vnd.ms-excel"
    format = "xls"


class ExportPDFTableRenderer(ExportFileRenderer):
    """Renders dataset as PDF (.pdf) in table format"""
    media_type = "application/pdf"
    format = "pdf_table"

    def export_set(self, formatted, headers, columns_per_page):
        stream = BytesIO()
        doc = SimpleDocTemplate(stream, pagesize=landscape(letter))
        styles = getSampleStyleSheet()
        styleCell = styles["Normal"]
        styleCell.fontSize = 7
        elements = []

        if headers:
            # slice the data into a set number of columns
            # in order to fit on the page
            for start in range(0, len(headers), columns_per_page):
                end = start + columns_per_page
                data = [['Row'] + [
                    item if item is not None else ''
                    for item in headers[start:end]
                ]]

                row_num = 1
                for row in formatted:
                    d = [
                        Paragraph(str(v), styleCell)
                        for _, v in row.items()
                    ]
                    data.append([row_num] + d[start:end])
                    row_num += 1

                t = Table(data)
                t.setStyle(TableStyle([
                    # action/format, from-cell, to-cell, format
                    ('VALIGN', (0, 0), (-1, -1), "TOP"),
                    ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
                    ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                ]))
                elements.append(t)
                elements.append(PageBreak())

        doc.build(elements)
        return stream.getvalue()

    def export_failure(self):
        stream = BytesIO()
        doc = SimpleDocTemplate(stream, pagesize=landscape(letter))
        styles = getSampleStyleSheet()
        styleCell = styles["Normal"]
        styleCell.fontSize = 7
        elements = [
            Paragraph(
                (
                    "Data not able to be formatted for PDF, "
                    "please try another format."
                ),
                styleCell
            )
        ]
        doc.build(elements)
        return stream.getvalue()

    def render_dataset(self, data, *args, **kwargs):
        formatted = data._package()
        headers = data.headers
        with open(self.filename, "wb") as fp:
            columns_per_page = PDF_COLUMNS_PER_PAGE
            success = False
            while columns_per_page > 1:
                try:
                    fp.write(
                        self.export_set(formatted, headers, columns_per_page)
                    )
                except LayoutError:
                    columns_per_page -= 1
                else:
                    success = True
                    break
            if not success:
                fp.write(self.export_failure())


class ExportPDFRenderer(ExportFileRenderer):
    """Renders dataset as PDF (.pdf)"""
    media_type = "application/pdf"
    format = "pdf"

    def export_set(self, formatted):
        stream = BytesIO()
        doc = SimpleDocTemplate(stream)
        styles = getSampleStyleSheet()
        styleParagraph = styles["Normal"]
        styleParagraph.fontSize = 7
        elements = []

        for row in formatted:
            if row:
                for k, val in row.items():
                    p = Paragraph(f"<b>{k}:</b> {val}", styleParagraph)
                    elements.append(p)
                elements.append(PageBreak())

        doc.build(elements)
        return stream.getvalue()

    def render_dataset(self, data, *args, **kwargs):
        formatted = data._package()
        with open(self.filename, "wb") as fp:
            fp.write(self.export_set(formatted))


class ExportDocxRenderer(ExportFileRenderer):
    """Renders dataset as Doc (.docx)"""
    media_type = "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    format = "docx"

    def export_set(self, formatted):
        stream = BytesIO()
        doc = Document()

        for row in formatted:
            if row:
                for k, val in row.items():
                    p = doc.add_paragraph()
                    p.add_run(f"{k}:").bold = True
                    p.add_run(f"{val}")
                doc.add_page_break()

        doc.save(stream)
        return stream.getvalue()

    def render_dataset(self, data, *args, **kwargs):
        formatted = data._package()
        with open(self.filename, "wb") as fp:
            fp.write(self.export_set(formatted))


class FriendlyCSVRenderer(CSVRenderer):
    def flatten_item(self, item):
        if isinstance(item, bool):
            return {'': {True: 'Yes', False: ''}[item]}
        return super().flatten_item(item)



from rest_framework import serializers
from tablib import Dataset


class ExportSerializer(serializers.ListSerializer):
    """Transforms data into a dataset"""
    read_only = True

    def get_header_label(self, field):
        try:
            label = self.child.fields[field].label
        except KeyError:
            label = field
        return label

    def get_headers(self, data):
        headers = []
        for field in data[0].keys():
            headers.append(self.get_header_label(field))
        return headers

    def get_dataset(self, data):
        headers = self.get_headers(data)
        data_list = []
        for d in data:
            data_list.append([v for _, v in d.items()])
        dataset = Dataset(*data_list, headers=headers)
        return dataset

    def transform_dataset(self, dataset):
        view = self.context.get('view', None)
        if view and hasattr(view, 'transform_dataset'):
            return self.context['view'].transform_dataset(dataset)
        return dataset

    @property
    def data(self):
        data = super(serializers.ListSerializer, self).data
        if isinstance(data, Dataset) or data:
            dataset = self.get_dataset(data)
            return self.transform_dataset(dataset)
        else:
            return Dataset([])



from django.conf import settings
from rest_framework.generics import ListAPIView
from rest_framework.mixins import ListModelMixin
from rest_framework.settings import perform_import
from rest_framework.viewsets import GenericViewSet

from unicef_rest_export.renderers import ExportBaseRenderer
from unicef_rest_export.serializers import ExportSerializer

DEFAULT_TEMPLATE = False
EXPORT_RENDERERS = getattr(settings, "EXPORT_RENDERERS", None)
if EXPORT_RENDERERS is None:
    EXPORT_RENDERERS = (
        "unicef_rest_export.renderers.ExportCSVRenderer",
        "unicef_rest_export.renderers.ExportJSONRenderer",
        "unicef_rest_export.renderers.ExportOpenXMLRenderer",
        "unicef_rest_export.renderers.ExportExcelRenderer",
        "unicef_rest_export.renderers.ExportPDFRenderer",
        "unicef_rest_export.renderers.ExportPDFTableRenderer",
        "unicef_rest_export.renderers.ExportDocxRenderer",
    )
    if "unicef_rest_export" in settings.INSTALLED_APPS:
        DEFAULT_TEMPLATE = True
        EXPORT_RENDERERS = (
            "unicef_rest_export.renderers.ExportHTMLRenderer",
        ) + EXPORT_RENDERERS

EXPORT_RENDERERS = perform_import(EXPORT_RENDERERS, "EXPORT_RENDERERS")


class ExportMixin(object):
    export_serializer_class = ExportSerializer

    def with_list_serializer(self, cls):
        meta = getattr(cls, 'Meta', object)
        if getattr(meta, 'list_serializer_class', None):
            return cls

        class SerializerWithListSerializer(cls):
            class Meta(meta):
                list_serializer_class = self.export_serializer_class

        return SerializerWithListSerializer

    def get_serializer_class(self):
        # c.f rest_framework.generics.GenericAPIView
        # (not using super() since this is a mixin class)
        assert self.serializer_class is not None, (
            "'%s' should either include a `serializer_class` attribute, "
            "or override the `get_serializer_class()` method."
            % self.__class__.__name__
        )

        renderer = self.request.accepted_renderer
        if isinstance(renderer, ExportBaseRenderer):
            return self.with_list_serializer(self.serializer_class)
        else:
            return self.serializer_class


class ExportViewBase(ExportMixin):
    renderer_classes = EXPORT_RENDERERS
    pagination_class = None
    if DEFAULT_TEMPLATE:
        template_name = 'rest_export.html'


class ExportView(ExportViewBase, ListAPIView):
    """Export-capable model list view"""
    pass


class ExportViewSet(ExportViewBase, ListModelMixin, GenericViewSet):
    """Export-capable model ViewSet (list only)"""
    pass


class ExportModelView(ListAPIView):
    """Label for export would be retrieved from label attribute"""

    @staticmethod
    def set_labels(serializer_fields, model):
        labels = {}
        model_labels = {}
        for f in model._meta.fields:
            model_labels[f.name] = f.verbose_name
        for f in serializer_fields:
            if model_labels.get(f, False):
                labels[f] = model_labels.get(f)
            elif serializer_fields.get(f) and serializer_fields[f].label:
                labels[f] = serializer_fields[f].label
            else:
                labels[f] = f.replace("_", " ").title()
        return labels

    def get_renderer_context(self):
        context = super().get_renderer_context()
        if hasattr(self, "get_serializer_class"):
            serializer_class = self.get_serializer_class()
            serializer = serializer_class()
            serializer_fields = serializer.get_fields()
            model = getattr(serializer.Meta, "model")
            context["labels"] = self.set_labels(
                serializer_fields,
                model
            )
        return context



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ast
import codecs
import os.path
import re
import subprocess
import sys
from codecs import open
from distutils import log
from distutils.errors import DistutilsError

from setuptools import find_packages, setup
from setuptools.command.install import install
from setuptools.command.sdist import sdist as BaseSDistCommand

ROOT = os.path.realpath(os.path.dirname(__file__))
init = os.path.join(ROOT, 'src', 'unicef_rest_export', '__init__.py')
_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

sys.path.insert(0, os.path.join(ROOT, 'src'))

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    VERSION = str(ast.literal_eval(_version_re.search(content).group(1)))
    NAME = str(ast.literal_eval(_name_re.search(content).group(1)))


def read(*files):
    content = []
    for f in files:
        content.extend(codecs.open(os.path.join(ROOT, 'src', 'requirements', f), 'r').readlines())
    return "\n".join(filter(lambda l:not l.startswith('-'), content))


def check(cmd, filename):
    out = subprocess.run(cmd, stdout=subprocess.PIPE)
    f = os.path.join('src', 'requirements', filename)
    reqs = codecs.open(os.path.join(ROOT, f), 'r').readlines()
    existing = {re.split("(==|>=|<=>|<|)", name[:-1])[0] for name in reqs}
    declared = {re.split("(==|>=|<=>|<|)", name)[0] for name in out.stdout.decode('utf8').split("\n") if name and not name.startswith('-')}

    if existing != declared:
        msg = """Requirements file not updated.
Run 'make requiremets'
""".format(' '.join(cmd), f)
        raise DistutilsError(msg)

class SDistCommand(BaseSDistCommand):

    def run(self):
        checks = {'install.pip': ['pipenv', 'lock', '--requirements'],
                  'testing.pip': ['pipenv', 'lock', '-d', '--requirements']}

        for filename, cmd in checks.items():
            check (cmd, filename)
        super().run()


class VerifyTagVersion(install):
    """Verify that the git tag matches version"""

    def run(self):
        tag = os.getenv("CIRCLE_TAG")
        if tag != VERSION:
            info = "Git tag: {} does not match the version of this app: {}".format(
                tag,
                VERSION
            )
            sys.exit(info)


setup(name=NAME,
      version=VERSION,
      url='https://github.com/unicef/unicef-rest-export',
      author='UNICEF',
      author_email='dev@unicef.org',
      license="Apache 2 License",
      description='Django package that handles exporting of data',
      long_description=codecs.open('README.md').read(),
      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      include_package_data=True,
      install_requires=read('install.pip'),
      extras_require={
          'test': read('install.pip', 'testing.pip'),
      },
      platforms=['any'],
      classifiers=[
          'Environment :: Web Environment',
          'Programming Language :: Python :: 3.6',
          'Framework :: Django',
          'Intended Audience :: Developers'],
      scripts=[],
      cmdclass={
          'sdist': SDistCommand,
          "verify": VerifyTagVersion,
      }
)
